from typing import  Dict, Mapping, Callable
from datasets import Dataset,concatenate_datasets
from cyvidia_ai_utils.src.cross_validation.cross_validation_model.cross_validation_model import CrossValidationModel
from cyvidia_ai_utils.src.cross_validation.evaluation_result import EvalAverageType, EvaluationResult
import pandas as pd
def cross_validate(
    create_model_instance: Callable[[],CrossValidationModel],
    folds: list[Dataset] | Mapping[str, Dataset],
    target_id_column:str,
    input_text_column:str,
    train_only_folds: list[Dataset]= [],
    average: EvalAverageType= EvalAverageType.WEIGHTED
):
    """Cross validate the model using the given folds and target column.
    Returns a dictionary of validation fold index to evaluation result.
    """
    if isinstance(folds, Mapping):
        fold_list: list[Dataset]=[]
        fold_labels:list[str]=[]
        
        for label, fold in folds.items():
            fold_list.append(fold)
            fold_labels.append(label)
    else:
        fold_list= folds
        fold_labels= [str(i) for i in range(len(folds))]
        
    results: Dict[str, EvaluationResult] = {}
    for i, val_ds in enumerate(fold_list):
        train_folds = fold_list[:i] + fold_list[i+1:]+ train_only_folds
        
        train_ds= concatenate_datasets(train_folds)
        
        model= create_model_instance()
        model.train(train_ds, val_ds)
        result= __evaluate_model(model, val_ds, target_column=target_id_column, input_column=input_text_column, average=average)
        print(f"Fold {fold_labels[i]} result: {result.to_json()}")
        results[fold_labels[i]] = result
    return results

def __evaluate_model(model: 'CrossValidationModel',test_ds_in: Dataset,input_column:str, target_column:str, average: EvalAverageType= EvalAverageType.WEIGHTED)-> EvaluationResult:
    test_df=test_ds_in.to_pandas()
    
    assert(isinstance(test_df, pd.DataFrame))
    test_df= test_df.drop_duplicates(subset=[input_column]).reset_index(drop=True)
    
    test_ds= Dataset.from_pandas(test_df, preserve_index=False)

    predictions_with_scores= model.predict_values(test_ds[input_column])

    predictions= [pred['label'] for pred in predictions_with_scores]
        
    true_values = list(map(lambda x: model.get_label_for_id(x), test_ds[target_column]))
    
    if len(true_values) != len(predictions):
        raise ValueError(f"The number of true values and predictions must be the same: {len(true_values)} != {len(predictions)}")

    if len(true_values)== 0:
        raise ValueError(f"The number of true values and predictions must be greater than 0. Length of test_ds: {len(test_ds)}")
    return EvaluationResult.from_values(true_values, predictions, average)          
        


from typing import Any, Dict
from cyvidia_ai_utils.src.cross_validation.cross_validation_model.cross_validation_model import CrossValidationModel
from transformers import PreTrainedModel, Trainer,PreTrainedTokenizer, pipeline,PreTrainedTokenizerFast, TrainingArguments,DataCollatorWithPadding

class TransformerCrossValidationModel(CrossValidationModel):
    def __init__(self, model: PreTrainedModel, tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast, training_args: TrainingArguments):
        self.model = model
        self.tokenizer = tokenizer
        self.training_args= training_args
        
    def get_label_for_id(self, id: int)-> str:
        return self.model.config.id2label[id]

    def train(self, train_ds, val_ds):
        def preprocess_function(examples):
            return self.tokenizer(examples["text"], truncation=True)
    
        tokenized_train_ds = train_ds.map(preprocess_function, batched=True)
        tokenized_val_ds = val_ds.map(preprocess_function, batched=True)
        data_collator = DataCollatorWithPadding(tokenizer=self.tokenizer)
        
        trainer = Trainer(
            model=self.model,
            args=self.training_args,
            train_dataset=tokenized_train_ds,
            eval_dataset=tokenized_val_ds,
            tokenizer=self.tokenizer,
            data_collator=data_collator,
        )
        
        trainer.train()
        
        self.model= trainer.model
        

    def predict_values(self, values)-> Dict[str,Any]:
        classifier= pipeline("text-classification", model=self.model, tokenizer=self.tokenizer)
        
        return classifier(values)



from abc import abstractmethod
from typing import Any, Dict
from datasets import Dataset


class CrossValidationModel:
    """Implement this interface to create a model that can be cross-validated."""
    @abstractmethod
    def train(self, train_ds: Dataset, val_ds:Dataset)-> 'CrossValidationModel':
        """Train the model using the given dataset and return the trained model."""
        pass
    
    @abstractmethod
    def predict_values(self, values)-> list[Dict[str,Any]]:
        '''Return the predicted values for the given input values.
        return looks like `[{"label": "POSITIVE", "score": 0.9999}, {"label": "NEGATIVE", "score": 0.7001}]`
        '''
        pass

    @abstractmethod
    def get_label_for_id(self, id: int)-> str:...
        


from cyvidia_ai_utils.src.cross_validation.cross_validation import cross_validate
from cyvidia_ai_utils.src.cross_validation.evaluation_result import EvaluationResult, EvalAverageType
from cyvidia_ai_utils.src.cross_validation.cross_validation_model.cross_validation_model import  CrossValidationModel
from cyvidia_ai_utils.src.cross_validation.cross_validation_model.transformer_cross_validation_model import TransformerCrossValidationModel


from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, Mapping, MutableMapping, Optional
from datasets import Dataset
import numpy as np
import json
import sklearn.metrics as metrics


class EvalAverageType(Enum):
    MICRO = "micro"
    MACRO = "macro"
    WEIGHTED = "weighted"
    BINARY= "binary"
    # SAMPLES= "samples"


class EvaluationResult:
    def __init__(
        self, 
        accuracy:float,
        f1_average: float,
        percision_average:float,
        per_label_f1: Mapping[str, float],
        per_label_percision: Mapping[str, float]
    ):
        self.accuracy = accuracy
        self.f1_average = f1_average
        self.percision_average = percision_average
        self.per_label_f1 = per_label_f1
        self.per_label_percision = per_label_percision
        
    @staticmethod
    def from_values(true_values, predictions, average: EvalAverageType):
        labels= list(set(true_values+ predictions))
        
        accuracy = metrics.accuracy_score(true_values, predictions)
        
        f1_average= metrics.f1_score(true_values, predictions, average=average.value)
        
        percision_average= metrics.precision_score(true_values, predictions, average=average.value)
        
        per_label_f1_list = metrics.f1_score(true_values, predictions, average=None, labels=labels)
        assert(isinstance(per_label_f1_list, np.ndarray))
        per_label_f1= dict(zip(labels, per_label_f1_list))
        
        per_label_percision_list = metrics.precision_score(true_values, predictions, average=None, labels=labels)
        assert(isinstance(per_label_percision_list, np.ndarray))
        per_label_percision= dict(zip(labels, per_label_percision_list))

        return EvaluationResult(
            accuracy=float(accuracy),
            f1_average=float(f1_average),
            percision_average=float(percision_average),
            per_label_f1=per_label_f1,
            per_label_percision=per_label_percision
        )
        
    def to_json(self)-> str:
        return json.dumps({
            "accuracy": self.accuracy,
            "f1_average": self.f1_average,
            "percision_average": self.percision_average,
            "per_label_f1": self.per_label_f1,
            "per_label_percision": self.per_label_percision
        })
    
    @staticmethod
    def aggregate(results: list['EvaluationResult'])-> 'EvaluationResult':
        accuracy= np.mean([result.accuracy for result in results])
        f1_average= np.mean([result.f1_average for result in results])
        percision_average= np.mean([result.percision_average for result in results])
        
        per_label_f1= {}
        per_label_percision= {}
        
        for result in results:
            for label, f1 in result.per_label_f1.items():
                if label in per_label_f1:
                    per_label_f1[label] += f1
                else:
                    per_label_f1[label] = f1
            for label, percision in result.per_label_percision.items():
                if label in per_label_percision:
                    per_label_percision[label] += percision
                else:
                    per_label_percision[label] = percision
        
        for label in per_label_f1:
            per_label_f1[label] /= len(results)
        for label in per_label_percision:
            per_label_percision[label] /= len(results)
        
        return EvaluationResult(
            accuracy=float(accuracy),
            f1_average=float(f1_average),
            percision_average=float(percision_average),
            per_label_f1=per_label_f1,
            per_label_percision=per_label_percision
        )

    def __str__(self):
        return f"Accuracy: {self.accuracy}, F1: {self.f1_average}, Percision: {self.percision_average}"
